import { Buffer as Buffer$1 } from 'buffer';
export { Buffer } from 'buffer';
import { WellKnownMimeType, encodeRoute, encodeBearerAuthMetadata, encodeSimpleAuthMetadata, encodeCompositeMetadata, encodeAndAddWellKnownMetadata, decodeCompositeMetadata, decodeRoutes } from '@rsocket/composite-metadata';
import * as i0 from '@angular/core';
import { signal, Injectable, inject } from '@angular/core';
import { ReplaySubject, defer, retry, timer, filter, first, map, switchMap, Subject, shareReplay } from 'rxjs';
import { deserializeFrame, FrameTypes, ErrorCodes, RSocketConnector } from '@rsocket/core';
import { WebsocketClientTransport } from '@rsocket/websocket-client';
import { nanoid } from 'nanoid';
import { toSignal } from '@angular/core/rxjs-interop';
import { print } from 'graphql/index';
import { OperationTypeNode } from 'graphql/language';

class RsocketConfig {
    constructor(rsocketServers, config = {}) {
        this.route = 'setup';
        this.rsocketServers = rsocketServers;
        this.config = config;
        this.payload({});
    }
    payload({ data, auth }) {
        if (!this.config.setup) {
            this.config['setup'] = {
                metadataMimeType: WellKnownMimeType.MESSAGE_RSOCKET_COMPOSITE_METADATA.string,
                dataMimeType: WellKnownMimeType.APPLICATION_JSON.string,
                keepAlive: 45000
            };
        }
        this.config.setup['payload'] = {
            data: data ? Buffer$1.from(JSON.stringify(data)) : undefined,
            metadata: this.buildAuthMetadata({ route: this.route, auth })
        };
        return this;
    }
    setAuth(auth) {
        if (this.config.setup?.payload?.metadata) {
            this.config.setup.payload.metadata = this.buildAuthMetadata({ route: this.route, auth });
        }
    }
    buildAuthMetadata({ route, auth }) {
        const map = new Map();
        map.set(WellKnownMimeType.MESSAGE_RSOCKET_ROUTING, encodeRoute(route));
        if (auth) {
            const { token, simple } = auth;
            if (token) {
                map.set(WellKnownMimeType.MESSAGE_RSOCKET_AUTHENTICATION, encodeBearerAuthMetadata(token));
            }
            else if (simple) {
                map.set(WellKnownMimeType.MESSAGE_RSOCKET_AUTHENTICATION, encodeSimpleAuthMetadata(simple.username, simple.password));
            }
        }
        return encodeCompositeMetadata(map);
    }
    getConfig() {
        return this.config;
    }
}

var ConnectionStatus;
(function (ConnectionStatus) {
    ConnectionStatus[ConnectionStatus["DISCONNECTED"] = 0] = "DISCONNECTED";
    ConnectionStatus[ConnectionStatus["REJECT"] = 1] = "REJECT";
    ConnectionStatus[ConnectionStatus["CONNECTING"] = 2] = "CONNECTING";
    ConnectionStatus[ConnectionStatus["CONNECTED"] = 3] = "CONNECTED";
})(ConnectionStatus || (ConnectionStatus = {}));

const websocketCreator = (url, serverName, abortSignal, { onOpen, onFirstFrame, onResumeOk, onRejectSetup, onResumeReject }) => {
    const websocket = new WebSocket(url);
    const originalSend = websocket.send;
    let resumeTimeout;
    websocket.send = (data) => {
        let frame = deserializeFrame(data);
        if (frame.type === FrameTypes.RESUME) {
            resumeTimeout = window.setTimeout(onResumeReject, 5000);
        }
        originalSend.call(websocket, data);
        websocket.send = originalSend;
    };
    function onFirstMessage(message) {
        onFirstFrame();
        const buffer = Buffer.from(message.data);
        let frame = deserializeFrame(buffer);
        switch (frame.type) {
            case FrameTypes.ERROR:
                switch (frame.code) {
                    case ErrorCodes.REJECTED_SETUP:
                        window.clearTimeout(resumeTimeout);
                        onRejectSetup();
                        break;
                    case ErrorCodes.REJECTED_RESUME:
                        window.clearTimeout(resumeTimeout);
                        onResumeReject();
                        break;
                }
                break;
            case FrameTypes.RESUME_OK:
                window.clearTimeout(resumeTimeout);
                onResumeOk();
                break;
        }
        websocket.removeEventListener('message', onFirstMessage);
    }
    const closeHandler = () => {
        websocket.removeEventListener('open', onOpen);
        websocket.removeEventListener('message', onFirstMessage);
        websocket.removeEventListener('close', closeHandler);
    };
    websocket.addEventListener('open', onOpen);
    websocket.addEventListener('message', onFirstMessage);
    websocket.addEventListener('close', closeHandler);
    abortSignal.addEventListener('abort', event => {
        if (sessionStorage.getItem('serverName') !== serverName) {
            websocket.close(1000);
        }
    });
    return websocket;
};

class RSocketClient {
    constructor(config, rsocketServer, waitFirstFrame, resolve, reject, abortSignal) {
        this.rSocket = new ReplaySubject(1);
        this.connectionStatus = new ReplaySubject(1);
        this.error = new ReplaySubject(1);
        this.inRacing = true;
        this.serverName = rsocketServer.name;
        let isRejectSetup = false;
        let abort = false;
        const connectorConfig = {
            ...config,
            transport: new WebsocketClientTransport({
                url: rsocketServer.url,
                wsCreator: (url) => websocketCreator(url, this.serverName, abortSignal, {
                    onOpen: () => {
                        this.connectionStatus.next(ConnectionStatus.CONNECTED);
                        if (this.inRacing && !waitFirstFrame) {
                            this.inRacing = false;
                            resolve(this);
                        }
                    },
                    onFirstFrame: () => {
                        if (this.inRacing && waitFirstFrame) {
                            this.inRacing = false;
                            resolve(this);
                        }
                    },
                    onResumeOk: () => {
                        isRejectSetup = false;
                    },
                    onResumeReject: () => {
                        isRejectSetup = false;
                        connector();
                    },
                    onRejectSetup: () => {
                        isRejectSetup = true;
                    }
                })
            }),
            resume: {
                tokenGenerator: () => Buffer$1.from(nanoid()),
                reconnectFunction: () => new Promise((resolve, reject) => {
                    this.connectionStatus.next(ConnectionStatus.DISCONNECTED);
                    if (isRejectSetup || abort) {
                        return reject();
                    }
                    console.log(`resume: ${rsocketServer.name} delay ${2000}`);
                    setTimeout(resolve, 2000);
                })
            }
        };
        const connector = () => defer(() => new RSocketConnector(connectorConfig).connect()).pipe(retry({ delay: () => timer(2000).pipe(filter(() => !abort)) })).subscribe({
            next: (rSocket) => {
                console.log(`Connected to ${this.serverName}`);
                rSocket.onClose(error => {
                    const rsocketError = error;
                    if (rsocketError) {
                        console.log(rsocketError);
                        this.error.next(rsocketError);
                        if (rsocketError.code === ErrorCodes.REJECTED_SETUP) {
                            this.connectionStatus.next(ConnectionStatus.REJECT);
                        }
                        else {
                            this.connectionStatus.next(ConnectionStatus.DISCONNECTED);
                        }
                    }
                });
                this.rSocket.next(rSocket);
                abortSignal.addEventListener('abort', () => {
                    if (abort) {
                        rSocket.close();
                    }
                });
            },
            error: (error) => {
                console.error('Failed to connect to RSocket server', error);
            }
        });
        abortSignal.addEventListener('abort', event => {
            if (sessionStorage.getItem('serverName') !== this.serverName) {
                abort = true;
                reject();
            }
        });
        connector();
    }
    encodePayload({ route, data, encodeData = true }) {
        let dataBuffer = null;
        if (data) {
            if (encodeData) {
                dataBuffer = Buffer$1.from(JSON.stringify(data));
            }
            else {
                dataBuffer = Buffer$1.from(data);
            }
        }
        return {
            data: dataBuffer,
            metadata: encodeAndAddWellKnownMetadata(Buffer$1.alloc(0), WellKnownMimeType.MESSAGE_RSOCKET_ROUTING, encodeRoute(route))
        };
    }
    fireAndForget({ payload }) {
        this.rSocket.pipe(first()).subscribe(rSocket => rSocket.fireAndForget(this.encodePayload(payload), {
            onError: (error) => {
            },
            onComplete: () => {
            }
        }));
    }
    requestResponse({ payload, responderStream }) {
        return this.rSocket.pipe(map(rSocket => rSocket.requestResponse(this.encodePayload(payload), {
            onError(error) {
                responderStream.onError?.(error);
            },
            onComplete: () => {
                responderStream.onComplete?.();
            },
            onNext(payload, isComplete) {
                responderStream.onNext(payload, isComplete);
            },
            onExtension(extendedType, content, canBeIgnored) {
                responderStream.onExtension?.(extendedType, content, canBeIgnored);
            }
        })));
    }
    requestStream({ payload, initialRequestN, responderStream }) {
        return this.rSocket.pipe(map(rSocket => {
            return rSocket.requestStream(this.encodePayload(payload), initialRequestN, {
                onError: (error) => {
                    responderStream.onError?.(error);
                },
                onNext(payload, isComplete) {
                    responderStream.onNext?.(payload, isComplete);
                },
                onComplete() {
                    responderStream.onComplete?.();
                },
                onExtension(extendedType, content, canBeIgnored) {
                    responderStream.onExtension?.(extendedType, content, canBeIgnored);
                }
            });
        }));
    }
    requestChannel({ payload, initialRequestN, isCompleted, responderStream }) {
        return this.rSocket.pipe(map(rSocket => rSocket.requestChannel(this.encodePayload(payload), initialRequestN, isCompleted, {
            onNext: (payload, isComplete) => {
                responderStream.onNext(payload, isComplete);
            },
            onComplete: () => {
                responderStream.onComplete?.();
            },
            onError(error) {
                responderStream.onError?.(error);
            },
            onExtension(extendedType, content, canBeIgnored) {
                responderStream.onExtension?.(extendedType, content, canBeIgnored);
            },
            request(requestN) {
                responderStream.request?.(requestN);
            },
            cancel() {
                responderStream.cancel?.();
            }
        })));
    }
}

var RequestType;
(function (RequestType) {
    RequestType[RequestType["REQUEST_FNF"] = 0] = "REQUEST_FNF";
    RequestType[RequestType["REQUEST_RESPONSE"] = 1] = "REQUEST_RESPONSE";
})(RequestType || (RequestType = {}));
class RsocketResponder {
    constructor() {
        this.routes = new Map();
    }
    route(path, handler) {
        this.routes.set(path, handler);
    }
    fireAndForget(payload, responderStream) {
        const handler = this.findHandler(payload.metadata);
        if (handler) {
            handler(payload.data ? JSON.parse(String(payload.data)) : undefined);
        }
        return this;
    }
    requestChannel(payload, initialRequestN, isCompleted, responderStream) {
        return this;
    }
    requestResponse(payload, responderStream) {
        const handler = this.findHandler(payload.metadata);
        if (handler) {
            responderStream.onNext({ data: Buffer$1.from(JSON.stringify(handler(payload.data ? JSON.parse(String(payload.data)) : undefined))) }, true);
        }
        return this;
    }
    requestStream(payload, initialRequestN, responderStream) {
        return this;
    }
    metadataPush(metadata, responderStream) {
    }
    cancel() {
    }
    request(requestN) {
    }
    onNext(payload, isComplete) {
    }
    onExtension(extendedType, content, canBeIgnored) {
    }
    onError(error) {
    }
    onComplete() {
    }
    close(error) {
    }
    onClose(callback) {
    }
    findHandler(metadata) {
        if (metadata && metadata.length) {
            for (let entry of decodeCompositeMetadata(metadata)) {
                if (entry.mimeType === WellKnownMimeType.MESSAGE_RSOCKET_ROUTING.string) {
                    for (let route of decodeRoutes(entry.content)) {
                        const handler = this.routes.get(route);
                        if (handler) {
                            return handler;
                        }
                    }
                }
            }
        }
        return undefined;
    }
}

class RsocketService {
    constructor() {
        this.rsocketClient = new ReplaySubject(1);
        this.connectionStatus = toSignal(this.rsocketClient.pipe(switchMap(rsocketClient => rsocketClient.connectionStatus)), { initialValue: ConnectionStatus.DISCONNECTED });
        this.error = toSignal(this.rsocketClient.pipe(switchMap(rsocketClient => rsocketClient.error)), { initialValue: undefined });
        this.serverLatency = signal({ name: '', latency: 0 });
        this.rsocketConfig = undefined;
        this.abortController = new AbortController();
        this.signal = this.abortController.signal;
        this.rSocketResponder = new RsocketResponder();
        this.responder = {
            fireAndForget: (route, handler) => {
                this.rSocketResponder.route(route, handler);
            },
            requestResponse: (route, handler) => {
                this.rSocketResponder.route(route, handler);
            }
        };
    }
    connect(waitFirstFrame = false) {
        const rsocketConfig = this.rsocketConfig;
        if (!rsocketConfig) {
            return;
        }
        rsocketConfig.getConfig()['responder'] = this.rSocketResponder;
        const start = performance.now();
        const promiseArray = rsocketConfig.rsocketServers.map(rsocketServer => new Promise((resolve, reject) => new RSocketClient(rsocketConfig.getConfig(), rsocketServer, waitFirstFrame, resolve, reject, this.signal)));
        Promise.race(promiseArray).then(rsocketClient => {
            console.log(rsocketClient.serverName, 'just won the race');
            this.rsocketClient.next(rsocketClient);
            this.serverLatency.set({ name: rsocketClient.serverName, latency: performance.now() - start });
            // TODO: chrome < 98, doesn't support abort reason, delete it after update
            sessionStorage.setItem('serverName', rsocketClient.serverName);
            this.abortController.abort();
            // this.abortController.abort(rsocketClient.serverName)
        });
    }
    fireAndForget(payload) {
        this.rsocketClient.pipe(first()).subscribe(rsocketClient => rsocketClient.fireAndForget(payload));
    }
    requestResponse(param) {
        return this.rsocketClient.pipe(switchMap(rsocketClient => rsocketClient.requestResponse(param)), first());
    }
    requestStream(param) {
        return this.rsocketClient.pipe(switchMap(rsocketClient => rsocketClient.requestStream(param)), first());
    }
    requestChannel(param) {
        return this.rsocketClient.pipe(switchMap(rsocketClient => rsocketClient.requestChannel(param)), first());
    }
    setConfig(config) {
        this.rsocketConfig = config;
    }
    setAuth(auth) {
        if (!this.rsocketConfig) {
            throw new Error("rsocketConfig is undefined.");
        }
        this.rsocketConfig.setAuth(auth);
    }
    static { this.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "18.2.3", ngImport: i0, type: RsocketService, deps: [], target: i0.ɵɵFactoryTarget.Injectable }); }
    static { this.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "18.2.3", ngImport: i0, type: RsocketService }); }
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "18.2.3", ngImport: i0, type: RsocketService, decorators: [{
            type: Injectable
        }] });

class GraphqlClient {
    constructor(rSocketClient, route) {
        this.rSocketClient = rSocketClient;
        this.route = route;
    }
    query(query) {
        const { documentNode, extract, variables } = query;
        const result = new Subject();
        return this.rSocketClient.requestResponse({
            payload: { data: { query: print(documentNode), variables }, route: this.route },
            responderStream: {
                onNext: (payload, isComplete) => {
                    result.next(extract(payload));
                },
                onComplete() {
                    result.complete();
                },
                onError(err) {
                    result.error(err);
                }
            }
        }).pipe(switchMap(() => result));
    }
    subscription(subscription, initialRequestN) {
        const { documentNode, extract, variables } = subscription;
        const result = new Subject();
        return this.rSocketClient.requestStream({
            payload: { data: { query: print(documentNode), variables }, route: this.route },
            initialRequestN,
            responderStream: {
                onNext: (payload, isComplete) => {
                    result.next(extract(payload));
                },
                onComplete: () => {
                    result.complete();
                },
                onError(err) {
                    result.error(err);
                }
            }
        }).pipe(map(requester => {
            return { result, requester };
        }));
    }
    mutation(mutation) {
        const { documentNode, extract, variables } = mutation;
        const result = new Subject();
        return this.rSocketClient.requestResponse({
            payload: { data: { query: print(documentNode), variables }, route: this.route },
            responderStream: {
                onNext: (payload, isComplete) => {
                    result.next(extract(payload));
                },
                onComplete() {
                    result.complete();
                },
                onError(err) {
                    result.error(err);
                }
            }
        }).pipe(switchMap(() => result));
    }
}

class GraphqlService {
    constructor() {
        this.rsocketService = inject(RsocketService);
        this.graphqlClient = this.rsocketService.rsocketClient.pipe(map(rsocketClient => new GraphqlClient(rsocketClient, 'graphql')), shareReplay(1));
    }
    query(query) {
        return this.graphqlClient.pipe(switchMap(graphqlClient => graphqlClient.query(query)), first());
    }
    subscription(subscription, initialRequestN = 0) {
        return this.graphqlClient.pipe(switchMap(graphqlClient => graphqlClient.subscription(subscription, initialRequestN)), first());
    }
    mutation(mutation) {
        return this.graphqlClient.pipe(switchMap(graphqlClient => graphqlClient.mutation(mutation)), first());
    }
    static { this.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "18.2.3", ngImport: i0, type: GraphqlService, deps: [], target: i0.ɵɵFactoryTarget.Injectable }); }
    static { this.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "18.2.3", ngImport: i0, type: GraphqlService }); }
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "18.2.3", ngImport: i0, type: GraphqlService, decorators: [{
            type: Injectable
        }] });

function getExtract(documentNode, operation) {
    const { definitions } = documentNode;
    if (definitions.length !== 1) {
        throw Error(`GraphqlExtractor: definitions.length != 1\n${print(documentNode)}`);
    }
    //@ts-ignore
    if (definitions.at(0).operation !== operation) {
        throw Error(`GraphqlExtractor: '${operation}' type error\n${print(documentNode)}`);
    }
    //@ts-ignore
    const { selections } = definitions.at(0).selectionSet;
    if (selections.length !== 1) {
        throw Error(`GraphqlExtractor: selections.length != 1\n${print(documentNode)}`);
    }
    return selections.at(0).name.value;
}
function create(documentNode, variables, operation) {
    const extract = getExtract(documentNode, operation);
    return {
        documentNode,
        variables,
        extract: (payload) => {
            if (!payload.data) {
                throw Error(`GraphqlExtractor => payload.data is null, query:\n${print(documentNode)}\nvariables:${variables}`);
            }
            const value = JSON.parse(payload.data.toString());
            if (!value['data']?.[extract] === undefined) {
                throw Error(`GraphqlExtractor => 'data' or '${extract}' not in ${JSON.stringify(value)}, document:`);
            }
            return value['data'][extract];
        }
    };
}
function createQuery(documentNode, variables) {
    return create(documentNode, variables, OperationTypeNode.QUERY);
}
function createSubscription(documentNode, variables) {
    return create(documentNode, variables, OperationTypeNode.SUBSCRIPTION);
}
function createMutation(documentNode, variables) {
    return create(documentNode, variables, OperationTypeNode.MUTATION);
}

/*
 * Public API Surface of ngx-rsocket-graphql
 */
window.Buffer = Buffer$1;

/**
 * Generated bundle index. Do not edit.
 */

export { ConnectionStatus, GraphqlService, RsocketConfig, RsocketService, createMutation, createQuery, createSubscription };
//# sourceMappingURL=ngx-rsocket-graphql.mjs.map
